/*
 * MySqlErd,  Entity Relation Diagram for MySQL databases.
 * Copyright (C) 2005, Michael McTernan, Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

import java.io.*;
import java.sql.*;
import java.util.regex.*;


/**
 * Program to create a ERD of some MySQL database.
 *  This uses foreign key relationships to render links between tables.
 *
 *@author Michael McTernan, Michael.McTernan.2001@cs.bris.ac.uk
 */
public class MySqlErd
{
    /**
     * Create a connection to the database.
     */
    static Connection createDbConnection(String host,
                                         String database,
                                         String user,
                                         String pass) throws ClassNotFoundException,
                                                             InstantiationException,
                                                             IllegalAccessException,
                                                             SQLException
    {
        /* Register the driver */
        Class.forName("com.mysql.jdbc.Driver").newInstance();

        /* Connect to the database */
        if(user != null && pass != null)
        {
            return DriverManager.getConnection("jdbc:mysql://" + host + "/" + database,
                                               user,
                                               pass);
        }
        else
        {
            return DriverManager.getConnection("jdbc:mysql://" + host + "/" + database);
        }

    }


    /**
     * Call a <CODE>TableProcessor</CODE> with the name of each table in some
     *  database.  This will execute a SHOW TABLES query to the server and then
     *  process the results, calling the <CODE>process</CODE> method of the
     *  passed <CODE>TableProcessor</CODE>.
     */
    public static void iterateTables(Connection     dbConnection,
                                     TableProcessor tableProcessor) throws SQLException
    {
        ResultSet tables;

        /* Create and execute statement to get the table names */
        tables = dbConnection.prepareStatement("SHOW TABLES;").executeQuery();


        if( !tables.first() )
        {
            System.err.println("No tables found.");
        }
        else
        {
            /* Iterate through each table name */
            do
            {
                tableProcessor.process(dbConnection, tables.getString(1));
            }
            while(tables.next());
        }
    }


    /**
     * Abstract class that needs to be extended to offer some functionality for
     *  processing table names.
     */
    private abstract static class TableProcessor
    {
        /**
         * Process some table.
         *@param dbConnection  The connection to the database.
         *@param tableName     The name of the table to process.
         */
        abstract void process(Connection dbConnection,
                              String     tableName) throws SQLException;
    }


    /**
     * <CODE>TableProcessor</CODE> that generates records for the tables.
     */
    public static class RecordGeneratingTableProcessor extends TableProcessor
    {

        void process(Connection dbConnection, String tableName) throws SQLException
        {
            ResultSet         columns;

            /* Create and execute statement to get the table names.
             *  Ideally a PreparedStatement would be used here, but that
             *  will place single quotes around tableName, causing the
             *  query to fail.
             */
            columns = dbConnection.prepareStatement("DESCRIBE " + tableName + ";").executeQuery();


            if( columns.first() )
            {
                /* Create the node in the output */
                System.out.print("  \"" + tableName +"\" [ shape=\"record\" label=\"\\n" + tableName + "\\n\\n" );


                /* Iterate through each column name */
                do
                {
                    String columnName = columns.getString("Field");
                    String keyType    = columns.getString("Key");

                    System.out.print("|<" + columnName + "> " + columnName);

                    if(keyType.compareTo("PRI") == 0)
                    {
                        System.out.print("(PK)");
                    }

                }
                while(columns.next());

                System.out.println("\"];");
            }
        }
    }


    /**
     * <CODE>TableProcessor</CODE> that generates records for the tables.
     */
    public static class RelationGeneratingTableProcessor extends TableProcessor
    {

        void process(Connection dbConnection, String tableName) throws SQLException
        {
            ResultSet creation;

            try
            {

                /* Get the statement that would create the table */
                creation = dbConnection.createStatement().executeQuery("SHOW CREATE TABLE " + tableName + ";");

                if( !creation.first() )
                {
                   System.err.println("No CREATE statement returned for '" + tableName + "'.");
                }
                else
                {
                    Pattern          fkPattern;
                    String           createStatement;
                    String           line;
                    LineNumberReader lnReader;

                    fkPattern = Pattern.compile("^ *CONSTRAINT `[\\w]+` FOREIGN KEY \\(`([\\w]+)`\\) REFERENCES `([\\w]+)` \\(`([\\w]+)`\\)( ON DELETE CASCADE)?,? *$");

                    createStatement = creation.getString("Create Table");
                    lnReader = new LineNumberReader( new StringReader( createStatement ) );

                    do
                    {
                        /* Read a line */
                        line = lnReader.readLine();

                        /* If not at EOF */
                        if(line != null)
                        {
                            /* Create a matcher for the pattern against the line */
                            Matcher matcher = fkPattern.matcher(line);

                            /* See if the matcher fits */
                            if(matcher.matches())
                            {
                                /* It matches, so a constraint can now be added */
                                if(matcher.group(4) != null)
                                {
                                    System.out.println("  \"" + tableName + "\":" + matcher.group(1) + "->\"" +
                                                       matcher.group(2) + "\":" + matcher.group(3) +
                                                       " [ arrowhead = \"odiamond\" ];");

                                }
                                else
                                {
                                    System.out.println("  \"" + tableName + "\":" + matcher.group(1) + "->\"" +
                                                       matcher.group(2) + "\":" + matcher.group(3) + ";");
                                }
                            }


                        }
                    } while ( line != null);

                }
            }
            catch(IOException ioe)
            {
                throw new RuntimeException(ioe);
            }
        }
    }


    public static void main(String args[]) throws Exception
    {
        CommandLineOptions clo;
        Connection         dbConnection;

		/* Check if the options file has been specified */
		if(args.length > 0 && (args[0].equals("-l") || args[0].equals("--licence")))
		{
			byte 		buf[] = new byte[1024];
			int	 		count;
			InputStream is;

            /* Open the licence text from the JAR */
			is = ClassLoader.getSystemClassLoader().getResourceAsStream("COPYING");

			do
			{
				count = is.read(buf);
				if(count >0 )
				{
					System.out.write(buf, 0, count);
				}
			}
			while(count > 0);

			is.close();
			System.exit(0);
		}

        /* Create a parser for the command line options */
        clo = new CommandLineOptions(args);
        clo.addOption("-D", "--database=", "database", " Database to use.");
        clo.addOption("-h", "--host=",     "host",     " Host to connect to.");
        clo.addOption("-p", "--password=", "password", " Password to use when connecting to server.");
        clo.addOption("-u", "--user=",     "user",     " User for login if not current user.");


        /* Check that the minimal options have been supplied */
        if(clo.get("host") == null || clo.get("database") == null)
        {
            System.out.println("Host and database name must be specified with the -D and -h options.");
            System.out.println("Usage: MySqlErd [options]\n");
            System.out.println(clo.usage());
            System.out.println("  --licence, -l\n  Display licence and exit.");
            System.out.println();
			System.out.println("MySqlErd version 0.1, Copyright (C) 2005 Michael C McTernan,");
			System.out.println("                                         Michael.McTernan.2001@cs.bris.ac.uk");
			System.out.println("MySqlErd comes with ABSOLUTELY NO WARRANTY.  This is free software, and you are");
			System.out.println("welcome to redistribute it under certain conditions; type");
			System.out.println("`java -jar MySqlErd.jar -l' for details.");

            System.out.println("\nMySqlErd written by Michael.McTernan.2001@cs.bris.ac.uk");
        }
        else
        {
            /* Create a connection to the database */
            dbConnection = createDbConnection(clo.get("host"),
                                              clo.get("database"),
                                              clo.get("user"),
                                              clo.get("password"));

            /* Start the graph */
            System.out.println("digraph G {");
            System.out.println("  graph [ rankdir = \"LR\" ];");
            System.out.println("  size=\"6.5625,9.28125\"; ratio=\"fill\"; center=\"true\"");

            /* Create records (nodes) then relations (edges) */
            iterateTables(dbConnection, new RecordGeneratingTableProcessor());
            iterateTables(dbConnection, new RelationGeneratingTableProcessor());

            /* End the graph */
            System.out.println("}");
        }
    }
}